@@ -21,10 +21,10 @@ package org.apache.felix.framework;
 import java.io.IOException;
 import java.net.InetAddress;
 import java.io.InputStream;
-import java.net.JarURLConnection;
 import java.net.URL;
 import java.net.URLConnection;
 import java.net.URLStreamHandler;
+import java.security.AccessControlException;
 import java.security.AllPermission;
 import java.util.ArrayList;
 import java.util.Enumeration;
@@ -34,30 +34,26 @@ import java.util.Iterator;
 import java.util.List;
 import java.util.Map;
 import java.util.NoSuchElementException;
-import java.util.Properties;
 import java.util.Set;
 
+import org.apache.felix.framework.Felix.FelixResolver;
+import org.apache.felix.framework.capabilityset.Attribute;
+import org.apache.felix.framework.capabilityset.Capability;
+import org.apache.felix.framework.capabilityset.Directive;
+import org.apache.felix.framework.resolver.Module;
 import org.apache.felix.framework.util.FelixConstants;
-import org.apache.felix.framework.util.SecurityManagerEx;
+import org.apache.felix.framework.util.StringMap;
 import org.apache.felix.framework.util.Util;
-import org.apache.felix.framework.util.manifestparser.Capability;
+import org.apache.felix.framework.util.manifestparser.CapabilityImpl;
 import org.apache.felix.framework.util.manifestparser.ManifestParser;
-import org.apache.felix.framework.util.manifestparser.R4Attribute;
-import org.apache.felix.framework.util.manifestparser.R4Directive;
-import org.apache.felix.framework.util.manifestparser.R4Library;
-import org.apache.felix.moduleloader.ICapability;
-import org.apache.felix.moduleloader.IContent;
-import org.apache.felix.moduleloader.IContentLoader;
-import org.apache.felix.moduleloader.IModuleDefinition;
-import org.apache.felix.moduleloader.IRequirement;
-import org.apache.felix.moduleloader.ISearchPolicy;
-import org.apache.felix.moduleloader.IURLPolicy;
+import org.apache.felix.framework.resolver.Content;
 import org.osgi.framework.AdminPermission;
 import org.osgi.framework.Bundle;
 import org.osgi.framework.BundleActivator;
 import org.osgi.framework.BundleContext;
 import org.osgi.framework.BundleException;
 import org.osgi.framework.Constants;
+import org.osgi.framework.Version;
 
 /**
  * The ExtensionManager class is used in several ways.
@@ -82,7 +78,7 @@ import org.osgi.framework.Constants;
 // with the parent classloader and one instance per framework instance that
 // keeps track of extension bundles and systembundle exports for that framework
 // instance.
-class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IContentLoader, IContent
+class ExtensionManager extends URLStreamHandler implements Content
 {
     // The private instance that is added to Felix.class.getClassLoader() -
     // will be null if extension bundles are not supported (i.e., we are not
@@ -91,9 +87,19 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
 
     static
     {
+        // pre-init the url sub-system as otherwise we don't work on gnu/classpath
+        ExtensionManager extensionManager = new ExtensionManager();
+        try 
+        {
+            (new URL("http://felix.extensions:9/")).openConnection();
+        }
+        catch (Throwable t)
+        {
+            // This doesn't matter much - we only need the above to init the url subsystem
+        }
+        
         // We use the secure action of Felix to add a new instance to the parent
         // classloader.
-        ExtensionManager extensionManager = new ExtensionManager();
         try
         {
             Felix.m_secureAction.addURLToURLClassLoader(Felix.m_secureAction.createURL(
@@ -101,7 +107,7 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
                 "http://felix.extensions:9/", extensionManager),
                 Felix.class.getClassLoader());
         }
-        catch (Exception ex)
+        catch (Throwable ex)
         {
             // extension bundles will not be supported.
             extensionManager = null;
@@ -109,14 +115,14 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
         m_extensionManager = extensionManager;
     }
 
-    private Logger m_logger = null;
-    private BundleInfo m_sbi = null;
-    private ICapability[] m_capabilities = null;
+    private final Logger m_logger;
+    private final Map m_headerMap = new StringMap(false);
+    private final Module m_systemBundleModule;
+    private List<Capability> m_capabilities = null;
     private Set m_exportNames = null;
-    private ISearchPolicy m_searchPolicy = null;
-    private IURLPolicy m_urlPolicy = null;
     private Object m_securityContext = null;
     private final List m_extensions;
+    private volatile Bundle[] m_extensionsCache;
     private final Set m_names;
     private final Map m_sourceToExtensions;
 
@@ -124,7 +130,10 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
     // classloader.
     private ExtensionManager()
     {
+        m_logger = null;
+        m_systemBundleModule = null;
         m_extensions = new ArrayList();
+        m_extensionsCache = new Bundle[0];
         m_names = new HashSet();
         m_sourceToExtensions = new HashMap();
     }
@@ -140,28 +149,25 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
      * @param config the configuration to read properties from.
      * @param systemBundleInfo the info to change if we need to add exports.
      */
-    ExtensionManager(Logger logger, Map configMap, BundleInfo sbi)
+    ExtensionManager(Logger logger, Felix felix)
     {
+        m_systemBundleModule = new ExtensionManagerModule(felix);
         m_extensions = null;
+        m_extensionsCache = null;
         m_names = null;
         m_sourceToExtensions = null;
         m_logger = logger;
-        m_sbi = sbi;
 
 // TODO: FRAMEWORK - Not all of this stuff really belongs here, probably only exports.
         // Populate system bundle header map.
-        // Note: This is a reference to the actual header map,
-        // so these changes are saved. Kind of hacky.
-        Map map = m_sbi.getCurrentHeader();
-        // Initialize header map as a case insensitive map.
-        map.put(FelixConstants.BUNDLE_VERSION,
-            configMap.get(FelixConstants.FELIX_VERSION_PROPERTY));
-        map.put(FelixConstants.BUNDLE_SYMBOLICNAME,
+        m_headerMap.put(FelixConstants.BUNDLE_VERSION,
+            felix.getConfig().get(FelixConstants.FELIX_VERSION_PROPERTY));
+        m_headerMap.put(FelixConstants.BUNDLE_SYMBOLICNAME,
             FelixConstants.SYSTEM_BUNDLE_SYMBOLICNAME);
-        map.put(FelixConstants.BUNDLE_NAME, "System Bundle");
-        map.put(FelixConstants.BUNDLE_DESCRIPTION,
+        m_headerMap.put(FelixConstants.BUNDLE_NAME, "System Bundle");
+        m_headerMap.put(FelixConstants.BUNDLE_DESCRIPTION,
             "This bundle is system specific; it implements various system services.");
-        map.put(FelixConstants.EXPORT_SERVICE,
+        m_headerMap.put(FelixConstants.EXPORT_SERVICE,
             "org.osgi.service.packageadmin.PackageAdmin," +
             "org.osgi.service.startlevel.StartLevel," +
             "org.osgi.service.url.URLHandlers");
@@ -171,65 +177,80 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
         // We must construct the system bundle's export metadata.
         // Get configuration property that specifies which class path
         // packages should be exported by the system bundle.
-        String syspkgs = (String) configMap.get(Constants.FRAMEWORK_SYSTEMPACKAGES);
+        String syspkgs = (String) felix.getConfig().get(FelixConstants.FRAMEWORK_SYSTEMPACKAGES);
         // If no system packages were specified, load our default value.
-        syspkgs = (syspkgs == null) ? loadDefaultSystemPackages(m_logger) : syspkgs;
+        syspkgs = (syspkgs == null)
+            ? Util.getDefaultProperty(logger, Constants.FRAMEWORK_SYSTEMPACKAGES)
+            : syspkgs;
+        syspkgs = (syspkgs == null) ? "" : syspkgs;
         // If any extra packages are specified, then append them.
-        String extra = (String) configMap.get(Constants.FRAMEWORK_SYSTEMPACKAGES_EXTRA);
+        String extra = (String) felix.getConfig().get(FelixConstants.FRAMEWORK_SYSTEMPACKAGES_EXTRA);
         syspkgs = (extra == null) ? syspkgs : syspkgs + "," + extra;
+        m_headerMap.put(FelixConstants.BUNDLE_MANIFESTVERSION, "2");
+        m_headerMap.put(FelixConstants.EXPORT_PACKAGE, syspkgs);
         try
         {
-            setCapabilities(
-                addModuleCapability(map,
-                    ManifestParser.parseExportHeader(syspkgs)));
+            ManifestParser mp = new ManifestParser(
+                m_logger, felix.getConfig(), m_systemBundleModule, m_headerMap);
+            List<Capability> caps = aliasSymbolicName(mp.getCapabilities());
+            setCapabilities(caps);
         }
         catch (Exception ex)
         {
-            m_capabilities = new ICapability[0];
+            m_capabilities = new ArrayList<Capability>(0);
             m_logger.log(
+                felix,
                 Logger.LOG_ERROR,
                 "Error parsing system bundle export statement: "
                 + syspkgs, ex);
         }
     }
 
-    private ICapability[] addModuleCapability(Map headerMap, ICapability[] caps)
+    private static List<Capability> aliasSymbolicName(List<Capability> caps)
     {
-        try
+        if (caps == null)
         {
-            // First parse the symbolic name header.
-            ICapability moduleCap = ManifestParser.parseBundleSymbolicName(headerMap);
-            // Create a copy of the module capability with the standard
-            // alias for the bundle symbolic name.
-            R4Directive[] dirs = ((Capability) moduleCap).getDirectives();
-            R4Attribute[] attrs = ((Capability) moduleCap).getAttributes();
-            R4Attribute[] attrsAlias = new R4Attribute[attrs.length];
-            System.arraycopy(attrs, 0, attrsAlias, 0, attrs.length);
-            // Modify the alias attributes to have the standard symbolic name.
-            for (int i = 0; i < attrsAlias.length; i++)
+            return new ArrayList<Capability>(0);
+        }
+
+        List<Capability> aliasCaps = new ArrayList<Capability>(caps);
+
+        for (int capIdx = 0; capIdx < aliasCaps.size(); capIdx++)
+        {
+            // Get the attributes and search for bundle symbolic name.
+            List<Attribute> attrs = aliasCaps.get(capIdx).getAttributes();
+            for (int i = 0; i < attrs.size(); i++)
             {
-                if (attrsAlias[i].getName().equalsIgnoreCase(Constants.BUNDLE_SYMBOLICNAME_ATTRIBUTE))
+                // If there is a bundle symbolic name attribute, add the
+                // standard alias as a value.
+                if (attrs.get(i).getName().equalsIgnoreCase(Constants.BUNDLE_SYMBOLICNAME_ATTRIBUTE))
                 {
-                    attrsAlias[i] = new R4Attribute(
-                        Constants.BUNDLE_SYMBOLICNAME_ATTRIBUTE, Constants.SYSTEM_BUNDLE_SYMBOLICNAME, false);
+                    // Make a copy of the attribute array.
+                    List<Attribute> aliasAttrs = new ArrayList<Attribute>(attrs);
+                    // Add the aliased value.
+                    aliasAttrs.set(i, new Attribute(
+                        Constants.BUNDLE_SYMBOLICNAME_ATTRIBUTE,
+                        new String[] {
+                            (String) attrs.get(i).getValue(), Constants.SYSTEM_BUNDLE_SYMBOLICNAME},
+                        false));
+                    // Create the aliased capability to replace the old capability.
+                    aliasCaps.set(capIdx, new CapabilityImpl(
+                        caps.get(capIdx).getModule(),
+                        caps.get(capIdx).getNamespace(),
+                        caps.get(capIdx).getDirectives(),
+                        aliasAttrs));
+                    // Continue with the next capability.
+                    break;
                 }
             }
-            // Create the alias capability.
-            ICapability aliasCap = new Capability(ICapability.MODULE_NAMESPACE, dirs, attrsAlias);
-            // Finally, add the module and alias capabilities to the
-            // existing capabilities.
-            ICapability[] temp = new ICapability[caps.length + 2];
-            System.arraycopy(caps, 0, temp, 2, caps.length);
-            temp[0] = moduleCap;
-            temp[1] = aliasCap;
-            return temp;
-        }
-        catch (BundleException ex)
-        {
-            // This should not happen, but in case it does, then just
-            // return the original array of capabilities.
-            return caps;
         }
+
+        return aliasCaps;
+    }
+
+    public Module getModule()
+    {
+        return m_systemBundleModule;
     }
 
     public synchronized Object getSecurityContext()
@@ -242,15 +263,6 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
         m_securityContext = securityContext;
     }
 
-    /**
-     * Check whether the given manifest headers are from an extension bundle.
-     */
-    boolean isExtensionBundle(Map headers)
-    {
-        return (ManifestParser.parseExtensionBundleHeader((String)
-                headers.get(Constants.FRAGMENT_HOST)) != null);
-    }
-
     /**
      * Add an extension bundle. The bundle will be added to the parent classloader
      * and it's exported packages will be added to the module definition
@@ -265,23 +277,31 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
      *          AdminPermission.EXTENSIONLIFECYCLE and security is enabled.
      * @throws Exception in case something goes wrong.
      */
-    void addExtensionBundle(Felix felix, FelixBundle bundle)
+    synchronized void addExtensionBundle(Felix felix, BundleImpl bundle)
         throws SecurityException, BundleException, Exception
     {
         Object sm = System.getSecurityManager();
         if (sm != null)
         {
-            ((SecurityManager) sm).checkPermission(
-                new AdminPermission(bundle, AdminPermission.EXTENSIONLIFECYCLE));
+            try
+            {
+                ((SecurityManager) sm).checkPermission(
+                    new AdminPermission(bundle, AdminPermission.EXTENSIONLIFECYCLE));
+            }
+            catch (SecurityException ex)
+            {
+                // TODO: SECURITY - we need to throw this exception because of the 4.2.0 ct
+                throw new AccessControlException(ex.getMessage());
+            }
         }
 
-        if (!((BundleProtectionDomain) bundle.getInfo().getProtectionDomain()).impliesDirect(new AllPermission()))
+        if (!((BundleProtectionDomain) bundle.getProtectionDomain()).impliesDirect(new AllPermission()))
         {
             throw new SecurityException("Extension Bundles must have AllPermission");
         }
 
-        R4Directive dir = ManifestParser.parseExtensionBundleHeader((String)
-            bundle.getInfo().getCurrentHeader().get(Constants.FRAGMENT_HOST));
+        Directive dir = ManifestParser.parseExtensionBundleHeader((String)
+            bundle.getCurrentModule().getHeaders().get(Constants.FRAGMENT_HOST));
 
         // We only support classpath extensions (not bootclasspath).
         if (!Constants.EXTENSION_FRAMEWORK.equals(dir.getValue()))
@@ -291,26 +311,25 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
                 "Unsupported Extension Bundle type!"));
         }
 
-        // Not sure whether this is a good place to do it but we need to lock
-        felix.acquireBundleLock(felix);
-
         try
         {
-            bundle.getInfo().setExtension(true);
-
             // Merge the exported packages with the exported packages of the systembundle.
-            ICapability[] exports = null;
+            List<Capability> exports = null;
             try
             {
-                exports = ManifestParser.parseExportHeader((String)
-                    bundle.getInfo().getCurrentHeader().get(Constants.EXPORT_PACKAGE));
+                exports = ManifestParser.parseExportHeader(
+                    m_logger, m_systemBundleModule,
+                    (String) bundle.getCurrentModule().getHeaders().get(Constants.EXPORT_PACKAGE),
+                    m_systemBundleModule.getSymbolicName(), m_systemBundleModule.getVersion());
+                exports = aliasSymbolicName(exports);
             }
             catch (Exception ex)
             {
                 m_logger.log(
+                    bundle,
                     Logger.LOG_ERROR,
                     "Error parsing extension bundle export statement: "
-                    + bundle.getInfo().getCurrentHeader().get(Constants.EXPORT_PACKAGE), ex);
+                    + bundle.getCurrentModule().getHeaders().get(Constants.EXPORT_PACKAGE), ex);
                 return;
             }
 
@@ -323,27 +342,22 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
             else
             {
                 // We don't support extensions (i.e., the parent is not an URLClassLoader).
-                m_logger.log(Logger.LOG_WARNING,
+                m_logger.log(bundle, Logger.LOG_WARNING,
                     "Unable to add extension bundle to FrameworkClassLoader - Maybe not an URLClassLoader?");
                 throw new UnsupportedOperationException(
                     "Unable to add extension bundle to FrameworkClassLoader - Maybe not an URLClassLoader?");
             }
-            ICapability[] temp = new ICapability[getCapabilities().length + exports.length];
-            System.arraycopy(getCapabilities(), 0, temp, 0, getCapabilities().length);
-            System.arraycopy(exports, 0, temp, getCapabilities().length, exports.length);
+            List<Capability> temp = new ArrayList<Capability>(m_capabilities.size() + exports.size());
+            temp.addAll(m_capabilities);
+            temp.addAll(exports);
             setCapabilities(temp);
         }
         catch (Exception ex)
         {
-            bundle.getInfo().setExtension(false);
             throw ex;
         }
-        finally
-        {
-            felix.releaseBundleLock(felix);
-        }
 
-        bundle.getInfo().setState(Bundle.RESOLVED);
+        felix.setBundleStateAndNotify(bundle, Bundle.RESOLVED);
     }
 
     /**
@@ -353,36 +367,36 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
      * @param felix the framework instance the extension bundle is installed in.
      * @param bundle the extension bundle to start if it has a Felix specific activator.
      */
-    void startExtensionBundle(Felix felix, FelixBundle bundle)
+    void startExtensionBundle(Felix felix, BundleImpl bundle)
     {
         String activatorClass = (String)
-        bundle.getInfo().getCurrentHeader().get(
+        bundle.getCurrentModule().getHeaders().get(
             FelixConstants.FELIX_EXTENSION_ACTIVATOR);
 
         if (activatorClass != null)
         {
             try
             {
+// TODO: SECURITY - Should this consider security?
                 BundleActivator activator = (BundleActivator)
                     felix.getClass().getClassLoader().loadClass(
                         activatorClass.trim()).newInstance();
 
-// TODO: KARL - This is kind of hacky, can we improve it?
+// TODO: EXTENSIONMANAGER - This is kind of hacky, can we improve it?
                 felix.m_activatorList.add(activator);
 
-                BundleContext context = m_sbi.getBundleContext();
+                BundleContext context = felix._getBundleContext();
 
-                bundle.getInfo().setBundleContext(context);
+                bundle.setBundleContext(context);
 
-                if (felix.getInfo().getState() == Bundle.ACTIVE)
+                if ((felix.getState() == Bundle.ACTIVE) || (felix.getState() == Bundle.STARTING))
                 {
                     Felix.m_secureAction.startActivator(activator, context);
                 }
             }
             catch (Throwable ex)
             {
-                ex.printStackTrace();
-                m_logger.log(Logger.LOG_WARNING,
+                m_logger.log(bundle, Logger.LOG_WARNING,
                     "Unable to start Felix Extension Activator", ex);
             }
         }
@@ -403,37 +417,10 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
         }
     }
 
-    //
-    // IModuleDefinition
-    //
-    public ICapability[] getCapabilities()
-    {
-        return m_capabilities;
-    }
-
-    void setCapabilities(ICapability[] capabilities)
+    private void setCapabilities(List<Capability> capabilities)
     {
         m_capabilities = capabilities;
-
-        // Note: This is a reference to the actual header map,
-        // so these changes are saved. Kind of hacky.
-        Map map = m_sbi.getCurrentHeader();
-        map.put(Constants.EXPORT_PACKAGE, convertCapabilitiesToHeaders(map));
-    }
-
-    public IRequirement[] getDynamicRequirements()
-    {
-        return null;
-    }
-
-    public R4Library[] getLibraries()
-    {
-        return null;
-    }
-
-    public IRequirement[] getRequirements()
-    {
-        return null;
+        m_headerMap.put(Constants.EXPORT_PACKAGE, convertCapabilitiesToHeaders(m_headerMap));
     }
 
     private String convertCapabilitiesToHeaders(Map headers)
@@ -441,9 +428,9 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
         StringBuffer exportSB = new StringBuffer("");
         Set exportNames = new HashSet();
 
-        for (int i = 0; (m_capabilities != null) && (i < m_capabilities.length); i++)
+        for (int i = 0; (m_capabilities != null) && (i < m_capabilities.size()); i++)
         {
-            if (m_capabilities[i].getNamespace().equals(ICapability.PACKAGE_NAMESPACE))
+            if (m_capabilities.get(i).getNamespace().equals(Capability.PACKAGE_NAMESPACE))
             {
                 // Add a comma separate if there is an existing package.
                 if (exportSB.length() > 0)
@@ -452,13 +439,31 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
                 }
 
                 // Append exported package information.
-                exportSB.append(((Capability) m_capabilities[i]).getPackageName());
-                exportSB.append("; version=\"");
-                exportSB.append(((Capability) m_capabilities[i]).getPackageVersion().toString());
-                exportSB.append("\"");
+                exportSB.append(m_capabilities.get(i).getAttribute(Capability.PACKAGE_ATTR).getValue());
+                for (Directive dir : m_capabilities.get(i).getDirectives())
+                {
+                    exportSB.append("; ");
+                    exportSB.append(dir.getName());
+                    exportSB.append(":=\"");
+                    exportSB.append(dir.getValue());
+                    exportSB.append("\"");
+                }
+                for (Attribute attr : m_capabilities.get(i).getAttributes())
+                {
+                    if (!attr.getName().equals(Capability.PACKAGE_ATTR)
+                        && !attr.getName().equals(Constants.BUNDLE_SYMBOLICNAME_ATTRIBUTE)
+                        && !attr.getName().equals(Constants.BUNDLE_VERSION_ATTRIBUTE))
+                    {
+                        exportSB.append("; ");
+                        exportSB.append(attr.getName());
+                        exportSB.append("=\"");
+                        exportSB.append(attr.getValue());
+                        exportSB.append("\"");
+                    }
+                }
 
                 // Remember exported packages.
-                exportNames.add(((Capability) m_capabilities[i]).getPackageName());
+                exportNames.add(m_capabilities.get(i).getAttribute(Capability.PACKAGE_ATTR).getValue());
             }
         }
 
@@ -467,105 +472,6 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
         return exportSB.toString();
     }
 
-    //
-    // IContentLoader
-    //
-
-    public void open()
-    {
-        // Nothing needed here.
-    }
-
-    public void close()
-    {
-        // Nothing needed here.
-    }
-
-    public IContent getContent()
-    {
-        return this;
-    }
-
-    public ISearchPolicy getSearchPolicy()
-    {
-        return m_searchPolicy;
-    }
-
-    public void setSearchPolicy(ISearchPolicy searchPolicy)
-    {
-        m_searchPolicy = searchPolicy;
-    }
-
-    public IURLPolicy getURLPolicy()
-    {
-        return m_urlPolicy;
-    }
-
-    public void setURLPolicy(IURLPolicy urlPolicy)
-    {
-        m_urlPolicy = urlPolicy;
-    }
-
-    public Class getClass(String name)
-    {
-        if (!m_exportNames.contains(Util.getClassPackage(name)))
-        {
-            return null;
-        }
-
-        try
-        {
-            return getClass().getClassLoader().loadClass(name);
-        }
-        catch (ClassNotFoundException ex)
-        {
-            m_logger.log(
-                Logger.LOG_WARNING,
-                ex.getMessage(),
-                ex);
-        }
-        return null;
-    }
-
-    public URL getResource(String name)
-    {
-        return getClass().getClassLoader().getResource(name);
-    }
-
-    public Enumeration getResources(String name)
-    {
-       try
-       {
-           return getClass().getClassLoader().getResources(name);
-       }
-       catch (IOException ex)
-       {
-           return null;
-       }
-    }
-
-    public URL getResourceFromContent(String name)
-    {
-        // There is no content for the system bundle, so return null.
-        return null;
-    }
-
-    public boolean hasInputStream(int index, String urlPath)
-    {
-        return (getClass().getClassLoader().getResource(urlPath) != null);
-    }
-
-    public InputStream getInputStream(int index, String urlPath)
-    {
-        return getClass().getClassLoader().getResourceAsStream(urlPath);
-    }
-
-    public String findLibrary(String name)
-    {
-        // No native libs associated with the system bundle.
-        return null;
-    }
-
     //
     // Classpath Extension
     //
@@ -574,7 +480,7 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
      * See whether any registered extension provides the class requested. If not
      * throw an IOException.
      */
-    protected synchronized URLConnection openConnection(URL url) throws IOException
+    public URLConnection openConnection(URL url) throws IOException
     {
         String path = url.getPath();
 
@@ -583,10 +489,18 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
             return new URLHandlersBundleURLConnection(url);
         }
 
-        for (Iterator iter = m_extensions.iterator(); iter.hasNext();)
+        Bundle[] extensions = m_extensionsCache;
+        URL result = null;
+        for (Bundle extBundle : extensions)
         {
-            URL result = ((FelixBundle) iter.next()).getInfo().getCurrentModule().getContentLoader().getResourceFromContent(path);
-
+            try
+            {
+                result = ((ModuleImpl) ((BundleImpl) extBundle).getCurrentModule()).getResourceLocal(path);
+            }
+            catch (Exception ex)
+            {
+                // Maybe the bundle went away, so ignore this exception.
+            }
             if (result != null)
             {
                 return result.openConnection();
@@ -615,6 +529,8 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
         sourceExtensions.add(extension);
 
         _add(extension.getSymbolicName(), extension);
+        m_extensionsCache = (Bundle[])
+                m_extensions.toArray(new Bundle[m_extensions.size()]);
     }
 
     private synchronized void _removeExtensions(Object source)
@@ -635,6 +551,8 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
                 Bundle bundle = (Bundle) extIter.next();
                 _add(bundle.getSymbolicName(), bundle);
             }
+            m_extensionsCache = (Bundle[])
+                m_extensions.toArray(new Bundle[m_extensions.size()]);
         }
     }
 
@@ -647,6 +565,11 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
         }
     }
 
+    public void close()
+    {
+        // Do nothing on close, since we have nothing open.
+    }
+
     public Enumeration getEntries()
     {
         return new Enumeration()
@@ -677,7 +600,7 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
         return null;
     }
 
-    public IContent getEntryAsContent(String name)
+    public Content getEntryAsContent(String name)
     {
         return null;
     }
@@ -691,50 +614,148 @@ class ExtensionManager extends URLStreamHandler implements IModuleDefinition, IC
     // Utility methods.
     //
 
-    /**
-     * The default name used for the default configuration properties file.
-    **/
-    public static final String DEFAULT_PROPERTIES_FILE_VALUE = "default.properties";
-
-    static String loadDefaultSystemPackages(Logger logger)
+    class ExtensionManagerModule extends ModuleImpl
     {
-        // If we cannot get configuration properties for any reason, then just
-        // attempt to load resource default.properties instead.
-        URL propURL = ExtensionManager.class.getClassLoader()
-            .getResource(DEFAULT_PROPERTIES_FILE_VALUE);
-        InputStream is = null;
-        try
+        private final Version m_version;
+        ExtensionManagerModule(Felix felix)
         {
-            // Load properties from URL.
-            is = propURL.openConnection().getInputStream();
-            Properties props = new Properties();
-            props.load(is);
-            is.close();
-            // Perform variable substitution for system properties.
-            for (Enumeration e = props.propertyNames(); e.hasMoreElements(); )
-            {
-                String name = (String) e.nextElement();
-                props.setProperty(name,
-                    Util.substVars(props.getProperty(name), name, null, props));
-            }
-            // Return system packages property.
-            return props.getProperty(Constants.FRAMEWORK_SYSTEMPACKAGES);
+            super(m_logger, felix.getConfig(), felix, "0",
+                felix.getBootPackages(), felix.getBootPackageWildcards());
+            m_version = new Version((String)
+                felix.getConfig().get(FelixConstants.FELIX_VERSION_PROPERTY));
         }
-        catch (Exception ex)
-        {
-            // Try to close input stream if we have one.
-            try
-            {
-                if (is != null) is.close();
-            }
-            catch (IOException ex2)
-            {
-                // Nothing we can do.
-            }
 
-            logger.log(
-                Logger.LOG_ERROR, "Unable to load any configuration properties.", ex);
-            return "";
+        public Map getHeaders()
+        {
+            synchronized (ExtensionManager.this)
+            {
+                return m_headerMap;
+            }
+        }
+
+        public List<Capability> getCapabilities()
+        {
+            synchronized (ExtensionManager.this)
+            {
+                return m_capabilities;
+            }
+        }
+
+        public String getSymbolicName()
+        {
+            return FelixConstants.SYSTEM_BUNDLE_SYMBOLICNAME;
+        }
+
+        public Version getVersion()
+        {
+            return m_version;
+        }
+
+        public Class getClassByDelegation(String name) throws ClassNotFoundException
+        {
+            Class clazz = null;
+            String pkgName = Util.getClassPackage(name);
+            if (shouldBootDelegate(pkgName))
+            {
+                try
+                {
+                    // Get the appropriate class loader for delegation.
+                    ClassLoader bdcl = getBootDelegationClassLoader();
+                    clazz = bdcl.loadClass(name);
+                    // If this is a java.* package, then always terminate the
+                    // search; otherwise, continue to look locally if not found.
+                    if (pkgName.startsWith("java.") || (clazz != null))
+                    {
+                        return clazz;
+                    }
+                }
+                catch (ClassNotFoundException ex)
+                {
+                    // If this is a java.* package, then always terminate the
+                    // search; otherwise, continue to look locally if not found.
+                    if (pkgName.startsWith("java."))
+                    {
+                        throw ex;
+                    }
+                }
+            }
+            if (clazz == null)
+            {
+                synchronized (ExtensionManager.this)
+                {
+                    if (!m_exportNames.contains(Util.getClassPackage(name)))
+                    {
+                        throw new ClassNotFoundException(name);
+                    }
+                }
+
+                clazz = getClass().getClassLoader().loadClass(name);
+            }
+            return clazz;
+        }
+
+        public URL getResourceByDelegation(String name)
+        {
+            return getClass().getClassLoader().getResource(name);
+        }
+
+        public Enumeration getResourcesByDelegation(String name)
+        {
+           try
+           {
+               return getClass().getClassLoader().getResources(name);
+           }
+           catch (IOException ex)
+           {
+               return null;
+           }
+        }
+
+        public Logger getLogger()
+        {
+            return m_logger;
+        }
+
+        public Map getConfig()
+        {
+            return null;
+        }
+
+        public FelixResolver getResolver()
+        {
+            return null;
+        }
+
+        public void attachFragmentContents(Content[] fragmentContents)
+            throws Exception
+        {
+            throw new UnsupportedOperationException("Should not be used!");
+        }
+
+        public void close()
+        {
+            // Nothing needed here.
+        }
+
+        public Content getContent()
+        {
+            return ExtensionManager.this;
+        }
+
+        public URL getEntry(String name)
+        {
+            // There is no content for the system bundle, so return null.
+            return null;
+        }
+
+        public boolean hasInputStream(int index, String urlPath)
+        {
+            return (getClass().getClassLoader().getResource(urlPath) != null);
+        }
+
+        public InputStream getInputStream(int index, String urlPath)
+        {
+            return getClass().getClassLoader().getResourceAsStream(urlPath);
         }
     }
-}
\ No newline at end of file
+}
