@@ -19,8 +19,8 @@
 package org.apache.felix.framework.cache;
 
 import java.io.*;
-import java.net.URLDecoder;
 
+import java.util.Map;
 import org.apache.felix.framework.Logger;
 import org.osgi.framework.Bundle;
 
@@ -69,7 +69,6 @@ public class BundleArchive
 
     private static final transient String BUNDLE_ID_FILE = "bundle.id";
     private static final transient String BUNDLE_LOCATION_FILE = "bundle.location";
-    private static final transient String CURRENT_LOCATION_FILE = "current.location";
     private static final transient String REVISION_LOCATION_FILE = "revision.location";
     private static final transient String BUNDLE_STATE_FILE = "bundle.state";
     private static final transient String BUNDLE_START_LEVEL_FILE = "bundle.startlevel";
@@ -78,14 +77,15 @@ public class BundleArchive
     private static final transient String REVISION_DIRECTORY = "version";
     private static final transient String DATA_DIRECTORY = "data";
     private static final transient String ACTIVE_STATE = "active";
+    private static final transient String STARTING_STATE = "starting";
     private static final transient String INSTALLED_STATE = "installed";
     private static final transient String UNINSTALLED_STATE = "uninstalled";
 
-    private Logger m_logger = null;
+    private final Logger m_logger;
+    private final Map m_configMap;
     private long m_id = -1;
-    private File m_archiveRootDir = null;
+    private final File m_archiveRootDir;
     private String m_originalLocation = null;
-    private String m_currentLocation = null;
     private int m_persistentState = -1;
     private int m_startLevel = -1;
     private long m_lastModified = -1;
@@ -101,6 +101,9 @@ public class BundleArchive
     **/
     public BundleArchive()
     {
+        m_logger = null;
+        m_configMap = null;
+        m_archiveRootDir = null;
     }
 
     /**
@@ -120,10 +123,11 @@ public class BundleArchive
      * @param is input stream from which to read the bundle content.
      * @throws Exception if any error occurs.
     **/
-    public BundleArchive(Logger logger, File archiveRootDir, long id,
+    public BundleArchive(Logger logger, Map configMap, File archiveRootDir, long id,
         String location, InputStream is) throws Exception
     {
         m_logger = logger;
+        m_configMap = configMap;
         m_archiveRootDir = archiveRootDir;
         m_id = id;
         if (m_id <= 0)
@@ -137,7 +141,7 @@ public class BundleArchive
         initialize();
 
         // Add a revision for the content.
-        revise(m_originalLocation, is);
+        revise(false, m_originalLocation, is);
     }
 
     /**
@@ -149,13 +153,14 @@ public class BundleArchive
      * </p>
      * @param logger the logger to be used by the archive.
      * @param archiveRootDir the archive root directory for storing state.
-     * @param id the bundle identifier associated with the archive.
+     * @param configMap configMap for BundleArchive
      * @throws Exception if any error occurs.
     **/
-    public BundleArchive(Logger logger, File archiveRootDir)
+    public BundleArchive(Logger logger, Map configMap, File archiveRootDir)
         throws Exception
     {
         m_logger = logger;
+        m_configMap = configMap;
         m_archiveRootDir = archiveRootDir;
 
         // Add a revision for each one that already exists in the file
@@ -195,7 +200,7 @@ public class BundleArchive
         // to read the location from the current revision - if that fails we
         // likely have an old bundle cache and read the location the old way.
         // The next revision will update the bundle cache.
-        revise(getRevisionLocation(revisionCount - 1), null);
+        revise(true, getRevisionLocation(revisionCount - 1), null);
     }
 
     /**
@@ -207,36 +212,34 @@ public class BundleArchive
     **/
     public synchronized long getId() throws Exception
     {
-        if (m_id > 0)
+        if (m_id <= 0)
         {
-            return m_id;
-        }
-
-        // Read bundle location.
-        InputStream is = null;
-        BufferedReader br = null;
-        try
-        {
-            is = BundleCache.getSecureAction()
-                .getFileInputStream(new File(m_archiveRootDir, BUNDLE_ID_FILE));
-            br = new BufferedReader(new InputStreamReader(is));
-            m_id = Long.parseLong(br.readLine());
-        }
-        catch (FileNotFoundException ex)
-        {
-            // HACK: Get the bundle identifier from the archive root directory
-            // name, which is of the form "bundle<id>" where <id> is the bundle
-            // identifier numbers. This is a hack to deal with old archives that
-            // did not save their bundle identifier, but instead had it passed
-            // into them. Eventually, this can be removed.
-            m_id = Long.parseLong(
-                m_archiveRootDir.getName().substring(
-                    BundleCache.BUNDLE_DIR_PREFIX.length()));
-        }
-        finally
-        {
-            if (br != null) br.close();
-            if (is != null) is.close();
+            // Read bundle location.
+            InputStream is = null;
+            BufferedReader br = null;
+            try
+            {
+                is = BundleCache.getSecureAction()
+                    .getFileInputStream(new File(m_archiveRootDir, BUNDLE_ID_FILE));
+                br = new BufferedReader(new InputStreamReader(is));
+                m_id = Long.parseLong(br.readLine());
+            }
+            catch (FileNotFoundException ex)
+            {
+                // HACK: Get the bundle identifier from the archive root directory
+                // name, which is of the form "bundle<id>" where <id> is the bundle
+                // identifier numbers. This is a hack to deal with old archives that
+                // did not save their bundle identifier, but instead had it passed
+                // into them. Eventually, this can be removed.
+                m_id = Long.parseLong(
+                    m_archiveRootDir.getName().substring(
+                        BundleCache.BUNDLE_DIR_PREFIX.length()));
+            }
+            finally
+            {
+                if (br != null) br.close();
+                if (is != null) is.close();
+            }
         }
 
         return m_id;
@@ -251,27 +254,25 @@ public class BundleArchive
     **/
     public synchronized String getLocation() throws Exception
     {
-        if (m_originalLocation != null)
+        if (m_originalLocation == null)
         {
-            return m_originalLocation;
-        }
-
-        // Read bundle location.
-        InputStream is = null;
-        BufferedReader br = null;
-        try
-        {
-            is = BundleCache.getSecureAction()
-                .getFileInputStream(new File(m_archiveRootDir, BUNDLE_LOCATION_FILE));
-            br = new BufferedReader(new InputStreamReader(is));
-            m_originalLocation = br.readLine();
-            return m_originalLocation;
-        }
-        finally
-        {
-            if (br != null) br.close();
-            if (is != null) is.close();
+            // Read bundle location.
+            InputStream is = null;
+            BufferedReader br = null;
+            try
+            {
+                is = BundleCache.getSecureAction()
+                    .getFileInputStream(new File(m_archiveRootDir, BUNDLE_LOCATION_FILE));
+                br = new BufferedReader(new InputStreamReader(is));
+                m_originalLocation = br.readLine();
+            }
+            finally
+            {
+                if (br != null) br.close();
+                if (is != null) is.close();
+            }
         }
+        return m_originalLocation;
     }
 
     /**
@@ -285,49 +286,53 @@ public class BundleArchive
     **/
     public synchronized int getPersistentState() throws Exception
     {
-        if (m_persistentState >= 0)
+        if (m_persistentState < 0)
         {
-            return m_persistentState;
-        }
+            // Get bundle state file.
+            File stateFile = new File(m_archiveRootDir, BUNDLE_STATE_FILE);
 
-        // Get bundle state file.
-        File stateFile = new File(m_archiveRootDir, BUNDLE_STATE_FILE);
-
-        // If the state file doesn't exist, then
-        // assume the bundle was installed.
-        if (!BundleCache.getSecureAction().fileExists(stateFile))
-        {
-            return Bundle.INSTALLED;
-        }
-
-        // Read the bundle state.
-        InputStream is = null;
-        BufferedReader br = null;
-        try
-        {
-            is = BundleCache.getSecureAction()
-                .getFileInputStream(stateFile);
-            br = new BufferedReader(new InputStreamReader(is));
-            String s = br.readLine();
-            if ((s != null) && s.equals(ACTIVE_STATE))
-            {
-                m_persistentState = Bundle.ACTIVE;
-            }
-            else if ((s != null) && s.equals(UNINSTALLED_STATE))
-            {
-                m_persistentState = Bundle.UNINSTALLED;
-            }
-            else
+            // If the state file doesn't exist, then
+            // assume the bundle was installed.
+            if (!BundleCache.getSecureAction().fileExists(stateFile))
             {
                 m_persistentState = Bundle.INSTALLED;
             }
-            return m_persistentState;
-        }
-        finally
-        {
-            if (br != null) br.close();
-            if (is != null) is.close();
+            else
+            {
+                // Read the bundle state.
+                InputStream is = null;
+                BufferedReader br = null;
+                try
+                {
+                    is = BundleCache.getSecureAction()
+                        .getFileInputStream(stateFile);
+                    br = new BufferedReader(new InputStreamReader(is));
+                    String s = br.readLine();
+                    if ((s != null) && s.equals(ACTIVE_STATE))
+                    {
+                        m_persistentState = Bundle.ACTIVE;
+                    }
+                    else if ((s != null) && s.equals(STARTING_STATE))
+                    {
+                        m_persistentState = Bundle.STARTING;
+                    }
+                    else if ((s != null) && s.equals(UNINSTALLED_STATE))
+                    {
+                        m_persistentState = Bundle.UNINSTALLED;
+                    }
+                    else
+                    {
+                        m_persistentState = Bundle.INSTALLED;
+                    }
+                }
+                finally
+                {
+                    if (br != null) br.close();
+                    if (is != null) is.close();
+                }
+            }
         }
+        return m_persistentState;
     }
 
     /**
@@ -344,38 +349,44 @@ public class BundleArchive
         // Write the bundle state.
         OutputStream os = null;
         BufferedWriter bw = null;
-        try
+        if (m_persistentState != state)
         {
-            os = BundleCache.getSecureAction()
-                .getFileOutputStream(new File(m_archiveRootDir, BUNDLE_STATE_FILE));
-            bw = new BufferedWriter(new OutputStreamWriter(os));
-            String s = null;
-            switch (state)
+            try
             {
-                case Bundle.ACTIVE:
-                    s = ACTIVE_STATE;
-                    break;
-                case Bundle.UNINSTALLED:
-                    s = UNINSTALLED_STATE;
-                    break;
-                default:
-                    s = INSTALLED_STATE;
-                    break;
+                os = BundleCache.getSecureAction()
+                    .getFileOutputStream(new File(m_archiveRootDir, BUNDLE_STATE_FILE));
+                bw = new BufferedWriter(new OutputStreamWriter(os));
+                String s = null;
+                switch (state)
+                {
+                    case Bundle.ACTIVE:
+                        s = ACTIVE_STATE;
+                        break;
+                    case Bundle.STARTING:
+                        s = STARTING_STATE;
+                        break;
+                    case Bundle.UNINSTALLED:
+                        s = UNINSTALLED_STATE;
+                        break;
+                    default:
+                        s = INSTALLED_STATE;
+                        break;
+                }
+                bw.write(s, 0, s.length());
+                m_persistentState = state;
+            }
+            catch (IOException ex)
+            {
+                m_logger.log(
+                    Logger.LOG_ERROR,
+                    getClass().getName() + ": Unable to record state - " + ex);
+                throw ex;
+            }
+            finally
+            {
+                if (bw != null) bw.close();
+                if (os != null) os.close();
             }
-            bw.write(s, 0, s.length());
-            m_persistentState = state;
-        }
-        catch (IOException ex)
-        {
-            m_logger.log(
-                Logger.LOG_ERROR,
-                getClass().getName() + ": Unable to record state - " + ex);
-            throw ex;
-        }
-        finally
-        {
-            if (bw != null) bw.close();
-            if (os != null) os.close();
         }
     }
 
@@ -388,37 +399,37 @@ public class BundleArchive
     **/
     public synchronized int getStartLevel() throws Exception
     {
-        if (m_startLevel >= 0)
+        if (m_startLevel < 0)
         {
-            return m_startLevel;
-        }
+            // Get bundle start level file.
+            File levelFile = new File(m_archiveRootDir, BUNDLE_START_LEVEL_FILE);
 
-        // Get bundle start level file.
-        File levelFile = new File(m_archiveRootDir, BUNDLE_START_LEVEL_FILE);
-
-        // If the start level file doesn't exist, then
-        // return an error.
-        if (!BundleCache.getSecureAction().fileExists(levelFile))
-        {
-            return -1;
-        }
-
-        // Read the bundle start level.
-        InputStream is = null;
-        BufferedReader br= null;
-        try
-        {
-            is = BundleCache.getSecureAction()
-                .getFileInputStream(levelFile);
-            br = new BufferedReader(new InputStreamReader(is));
-            m_startLevel = Integer.parseInt(br.readLine());
-            return m_startLevel;
-        }
-        finally
-        {
-            if (br != null) br.close();
-            if (is != null) is.close();
+            // If the start level file doesn't exist, then
+            // return an error.
+            if (!BundleCache.getSecureAction().fileExists(levelFile))
+            {
+                m_startLevel = -1;
+            }
+            else
+            {
+                // Read the bundle start level.
+                InputStream is = null;
+                BufferedReader br= null;
+                try
+                {
+                    is = BundleCache.getSecureAction()
+                        .getFileInputStream(levelFile);
+                    br = new BufferedReader(new InputStreamReader(is));
+                    m_startLevel = Integer.parseInt(br.readLine());
+                }
+                finally
+                {
+                    if (br != null) br.close();
+                    if (is != null) is.close();
+                }
+            }
         }
+        return m_startLevel;
     }
 
     /**
@@ -465,36 +476,36 @@ public class BundleArchive
     **/
     public synchronized long getLastModified() throws Exception
     {
-        if (m_lastModified >= 0)
+        if (m_lastModified < 0)
         {
-            return m_lastModified;
-        }
+            // Get bundle last modification time file.
+            File lastModFile = new File(m_archiveRootDir, BUNDLE_LASTMODIFIED_FILE);
 
-        // Get bundle last modification time file.
-        File lastModFile = new File(m_archiveRootDir, BUNDLE_LASTMODIFIED_FILE);
-
-        // If the last modification file doesn't exist, then
-        // return an error.
-        if (!BundleCache.getSecureAction().fileExists(lastModFile))
-        {
-            return 0;
-        }
-
-        // Read the bundle start level.
-        InputStream is = null;
-        BufferedReader br= null;
-        try
-        {
-            is = BundleCache.getSecureAction().getFileInputStream(lastModFile);
-            br = new BufferedReader(new InputStreamReader(is));
-            m_lastModified = Long.parseLong(br.readLine());
-            return m_lastModified;
-        }
-        finally
-        {
-            if (br != null) br.close();
-            if (is != null) is.close();
+            // If the last modification file doesn't exist, then
+            // return an error.
+            if (!BundleCache.getSecureAction().fileExists(lastModFile))
+            {
+                m_lastModified = 0;
+            }
+            else
+            {
+                // Read the bundle start level.
+                InputStream is = null;
+                BufferedReader br= null;
+                try
+                {
+                    is = BundleCache.getSecureAction().getFileInputStream(lastModFile);
+                    br = new BufferedReader(new InputStreamReader(is));
+                    m_lastModified = Long.parseLong(br.readLine());
+                }
+                finally
+                {
+                    if (br != null) br.close();
+                    if (is != null) is.close();
+                }
+            }
         }
+        return m_lastModified;
     }
 
     /**
@@ -547,9 +558,15 @@ public class BundleArchive
     {
         // Do some sanity checking.
         if ((fileName.length() > 0) && (fileName.charAt(0) == File.separatorChar))
-            throw new IllegalArgumentException("The data file path must be relative, not absolute.");
+        {
+            throw new IllegalArgumentException(
+                "The data file path must be relative, not absolute.");
+        }
         else if (fileName.indexOf("..") >= 0)
-            throw new IllegalArgumentException("The data file path cannot contain a reference to the \"..\" directory.");
+        {
+            throw new IllegalArgumentException(
+                "The data file path cannot contain a reference to the \"..\" directory.");
+        }
 
         // Get bundle data directory.
         File dataDir = new File(m_archiveRootDir, DATA_DIRECTORY);
@@ -600,7 +617,7 @@ public class BundleArchive
      * @param location the location string associated with the revision.
      * @throws Exception if any error occurs.
     **/
-    public synchronized void revise(String location, InputStream is)
+    public synchronized void revise(boolean isReload, String location, InputStream is)
         throws Exception
     {
         // If we have an input stream, then we have to use it
@@ -617,7 +634,10 @@ public class BundleArchive
             throw new Exception("Unable to revise archive.");
         }
 
-        setRevisionLocation(location, (m_revisions == null) ? 0 : m_revisions.length);
+        if (!isReload)
+        {
+            setRevisionLocation(location, (m_revisions == null) ? 0 : m_revisions.length);
+        }
 
         // Add new revision to revision array.
         if (m_revisions == null)
@@ -653,11 +673,9 @@ public class BundleArchive
             return false;
         }
 
-        String location = getRevisionLocation(m_revisions.length - 2);
-
         try
         {
-            m_revisions[m_revisions.length - 1].dispose();
+            m_revisions[m_revisions.length - 1].close();
         }
         catch(Exception ex)
         {
@@ -721,6 +739,51 @@ public class BundleArchive
         }
     }
 
+    public synchronized void close()
+    {
+        // Get the current revision count.
+        int count = getRevisionCount();
+        for (int i = 0; i < count; i++)
+        {
+            // Dispose of the revision, but this might be null in certain
+            // circumstances, such as if this bundle archive was created
+            // for an existing bundle that was updated, but not refreshed
+            // due to a system crash; see the constructor code for details.
+            if (m_revisions[i] != null)
+            {
+                try
+                {
+                    m_revisions[i].close();
+                }
+                catch (Exception ex)
+                {
+                    m_logger.log(
+                        Logger.LOG_ERROR,
+                            "Unable to close revision - "
+                            + m_revisions[i].getRevisionRootDir(), ex);
+                }
+            }
+        }
+    }
+
+    /**
+     * <p>
+     * This method closes any revisions and deletes the bundle archive directory.
+     * </p>
+     * @throws Exception if any error occurs.
+    **/
+    public synchronized void closeAndDelete()
+    {
+        // Close the revisions and delete the archive directory.
+        close();
+        if (!BundleCache.deleteDirectoryTree(m_archiveRootDir))
+        {
+            m_logger.log(
+                Logger.LOG_ERROR,
+                "Unable to delete archive directory - " + m_archiveRootDir);
+        }
+    }
+
     /**
      * <p>
      * This method removes all old revisions associated with the archive
@@ -730,23 +793,15 @@ public class BundleArchive
     **/
     public synchronized void purge() throws Exception
     {
-        // Get the current refresh count.
+        // Close the revisions and then delete all but the current revision.
+        // We don't delete it the current revision, because we want to rename it
+        // to the new refresh level.
+        close();
         long refreshCount = getRefreshCount();
-        // Get the current revision count.
         int count = getRevisionCount();
-
-        // Dispose and delete all but the current revision.
         File revisionDir = null;
         for (int i = 0; i < count - 1; i++)
         {
-            // Dispose of the revision, but this might be null in certain
-            // circumstances, such as if this bundle archive was created
-            // for an existing bundle that was updated, but not refreshed
-            // due to a system crash; see the constructor code for details.
-            if (m_revisions[i] != null)
-            {
-                m_revisions[i].dispose();
-            }
             revisionDir = new File(m_archiveRootDir, REVISION_DIRECTORY + refreshCount + "." + i);
             if (BundleCache.getSecureAction().fileExists(revisionDir))
             {
@@ -754,11 +809,6 @@ public class BundleArchive
             }
         }
 
-        // We still need to dispose the current revision, but we
-        // don't want to delete it, because we want to rename it
-        // to the new refresh level.
-        m_revisions[count - 1].dispose();
-
         // Save the current revision location for use later when
         // we recreate the revision.
         String location = getRevisionLocation(count -1);
@@ -780,24 +830,6 @@ public class BundleArchive
         m_revisions = new BundleRevision[] { revision };
     }
 
-    /**
-     * <p>
-     * This method disposes removes the bundle archive directory.
-     * </p>
-     * @throws Exception if any error occurs.
-    **/
-    /* package */ void dispose() throws Exception
-    {
-        if (!BundleCache.deleteDirectoryTree(m_archiveRootDir))
-        {
-            m_logger.log(
-                Logger.LOG_ERROR,
-                getClass().getName()
-                    + ": Unable to delete archive directory - "
-                    + m_archiveRootDir);
-        }
-    }
-
     /**
      * <p>
      * Initializes the bundle archive object by creating the archive
@@ -849,76 +881,6 @@ public class BundleArchive
         }
     }
 
-    /**
-     * <p>
-     * Returns the current location associated with the bundle archive,
-     * which is the last location from which the bundle was updated. It is
-     * necessary to keep track of this so it is possible to determine what
-     * kind of revision needs to be created when recreating revisions when
-     * the framework restarts.
-     * </p>
-     * @return the last update location.
-     * @throws Exception if any error occurs.
-    **/
-    private String getCurrentLocation() throws Exception
-    {
-        if (m_currentLocation != null)
-        {
-            return m_currentLocation;
-        }
-
-        // Read current location.
-        InputStream is = null;
-        BufferedReader br = null;
-        try
-        {
-            is = BundleCache.getSecureAction()
-                .getFileInputStream(new File(m_archiveRootDir, CURRENT_LOCATION_FILE));
-            br = new BufferedReader(new InputStreamReader(is));
-            m_currentLocation = br.readLine();
-            return m_currentLocation;
-        }
-        catch (FileNotFoundException ex)
-        {
-            return getLocation();
-        }
-        finally
-        {
-            if (br != null) br.close();
-            if (is != null) is.close();
-        }
-    }
-
-    /**
-     * <p>
-     * Set the current location associated with the bundle archive,
-     * which is the last location from which the bundle was updated. It is
-     * necessary to keep track of this so it is possible to determine what
-     * kind of revision needs to be created when recreating revisions when
-     * the framework restarts.
-     * </p>
-     * @throws Exception if any error occurs.
-    **/
-    private void setCurrentLocation(String location) throws Exception
-    {
-        // Save current location.
-        OutputStream os = null;
-        BufferedWriter bw = null;
-        try
-        {
-            os = BundleCache.getSecureAction()
-                .getFileOutputStream(new File(m_archiveRootDir, CURRENT_LOCATION_FILE));
-            bw = new BufferedWriter(new OutputStreamWriter(os));
-            bw.write(location, 0, location.length());
-            m_currentLocation = location;
-        }
-        finally
-        {
-            if (bw != null) bw.close();
-            if (os != null) os.close();
-        }
-    }
-
     /**
      * <p>
      * Creates a revision based on the location string and/or input stream.
@@ -954,7 +916,7 @@ public class BundleArchive
                 }
 
                 // Decode any URL escaped sequences.
-                location = URLDecoder.decode(location);
+                location = decode(location);
 
                 // Make sure the referenced file exists.
                 File file = new File(location.substring(FILE_PROTOCOL.length()));
@@ -968,22 +930,26 @@ public class BundleArchive
                 // flag set to true.
                 if (BundleCache.getSecureAction().isFileDirectory(file))
                 {
-                    result = new DirectoryRevision(m_logger, revisionRootDir, location);
+                    result = new DirectoryRevision(m_logger, m_configMap,
+                        revisionRootDir, location);
                 }
                 else
                 {
-                    result = new JarRevision(m_logger, revisionRootDir, location, true);
+                    result = new JarRevision(m_logger, m_configMap, revisionRootDir,
+                        location, true);
                 }
             }
             else if (location.startsWith(INPUTSTREAM_PROTOCOL))
             {
                 // Assume all input streams point to JAR files.
-                result = new JarRevision(m_logger, revisionRootDir, location, false, is);
+                result = new JarRevision(m_logger, m_configMap, revisionRootDir,
+                    location, false, is);
             }
             else
             {
                 // Anything else is assumed to be a URL to a JAR file.
-                result = new JarRevision(m_logger, revisionRootDir, location, false);
+                result = new JarRevision(m_logger, m_configMap, revisionRootDir,
+                    location, false);
             }
         }
         catch (Exception ex)
@@ -1005,6 +971,47 @@ public class BundleArchive
         return result;
     }
 
+    // Method from Harmony java.net.URIEncoderDecoder (luni subproject)
+    // used by URI to decode uri components.
+    private static String decode(String s) throws UnsupportedEncodingException 
+    {
+        StringBuffer result = new StringBuffer();
+        ByteArrayOutputStream out = new ByteArrayOutputStream();
+        for (int i = 0; i < s.length();) 
+        {
+            char c = s.charAt(i);
+            if (c == '%') 
+            {
+                out.reset();
+                do 
+                {
+                    if ((i + 2) >= s.length()) 
+                    {
+                        throw new IllegalArgumentException(
+                            "Incomplete % sequence at: " + i);
+                    }
+                    int d1 = Character.digit(s.charAt(i + 1), 16);
+                    int d2 = Character.digit(s.charAt(i + 2), 16);
+                    if ((d1 == -1) || (d2 == -1))
+                    {
+                        throw new IllegalArgumentException("Invalid % sequence ("
+                            + s.substring(i, i + 3)
+                            + ") at: " + String.valueOf(i));
+                    }
+                    out.write((byte) ((d1 << 4) + d2));
+                    i += 3;
+                }
+                while ((i < s.length()) && (s.charAt(i) == '%'));
+                result.append(out.toString("UTF-8"));
+                continue;
+            }
+            result.append(c);
+            i++;
+        }
+        return result.toString();
+    }
+
+
     /**
      * This utility method is used to retrieve the current refresh
      * counter value for the bundle. This value is used when generating
@@ -1021,39 +1028,40 @@ public class BundleArchive
     **/
     private long getRefreshCount() throws Exception
     {
-        // If we have already read the refresh counter file,
-        // then just return the result.
-        if (m_refreshCount >= 0)
+        // If the refresh counter is not yet initialized, do so now.
+        if (m_refreshCount < 0)
         {
-            return m_refreshCount;
+            // Get refresh counter file.
+            File counterFile = new File(m_archiveRootDir, REFRESH_COUNTER_FILE);
+
+            // If the refresh counter file doesn't exist, then
+            // assume the counter is at zero.
+            if (!BundleCache.getSecureAction().fileExists(counterFile))
+            {
+                m_refreshCount = 0;
+            }
+            else
+            {
+                // Read the bundle refresh counter.
+                InputStream is = null;
+                BufferedReader br = null;
+                try
+                {
+                    is = BundleCache.getSecureAction()
+                        .getFileInputStream(counterFile);
+                    br = new BufferedReader(new InputStreamReader(is));
+                    long counter = Long.parseLong(br.readLine());
+                    return counter;
+                }
+                finally
+                {
+                    if (br != null) br.close();
+                    if (is != null) is.close();
+                }
+            }
         }
 
-        // Get refresh counter file.
-        File counterFile = new File(m_archiveRootDir, REFRESH_COUNTER_FILE);
-
-        // If the refresh counter file doesn't exist, then
-        // assume the counter is at zero.
-        if (!BundleCache.getSecureAction().fileExists(counterFile))
-        {
-            return 0;
-        }
-
-        // Read the bundle refresh counter.
-        InputStream is = null;
-        BufferedReader br = null;
-        try
-        {
-            is = BundleCache.getSecureAction()
-                .getFileInputStream(counterFile);
-            br = new BufferedReader(new InputStreamReader(is));
-            long counter = Long.parseLong(br.readLine());
-            return counter;
-        }
-        finally
-        {
-            if (br != null) br.close();
-            if (is != null) is.close();
-        }
+        return m_refreshCount;
     }
 
     /**
@@ -1101,4 +1109,4 @@ public class BundleArchive
             if (os != null) os.close();
         }
     }
-}
+}
\ No newline at end of file
